% This is a test file that computes the entire Regularization path 
% for matrix completion under trace norm penalty.
%
% \lambda is a regularization parameter that belongs to a grid of values \{\lambda_min,..., \lambda_max \}.
% 
% We solve a problem of the type "minimize f(X) + \lambda ||X||_*" for a grid of
% \lambda values  using a first-order predictor-corrector approach.
%
% Refer "Low-rank optimization with trace norm penalty"
% Technical report (arXiv 1112.2318), 2011 for further details.
% Authors:
% Bamdev Mishra and Gilles Meyer
% {b.mishra, g.meyer}@ulg.ac.be

clear all; close all; clc;
RandStream.setDefaultStream(RandStream('mt19937ar','seed',sum(100*clock)));

addpath('./common');
addpath('./manifold');


%% Setup
d1 = 1000;
d2 = 1000;
r = 10;
f = 0.1;% fraction of known entries

%% Optimization algorithm is the trust-region algorithm
optim_algo = @tr_matrix_completion;

%% Paramters
params.tol = 1e-10;
params.vtol = 1e-10;
params.dg_tol = 1e-3;
params.dg_vtol = 1e-5;
params.smax_tol = 1e-3;
params.max_iter = 10000;
params.max_iter_tr = 100;
params.verb = true;

%% Target model
Us = randn(d1,r);
Vs = randn(d2,r);
n  = d1 * d2;

%% Partition
[train, test] = crossvalind('HoldOut',n,1-f);  

%%% If there is no crossvalind then uncomment the following two lines

% test =(rand(n,1) > f); % w is a vector with fraction f of zeros
% train = logical(ones(n,1) - test);


[data_ls.rows, data_ls.cols] = ind2sub([d1,d2],find(train)); % training set
[data_ts.rows, data_ts.cols] = ind2sub([d1,d2],find(test)); % testing set

data_ls.entries = proj_omega_sub(Us,Vs,data_ls.rows,data_ls.cols); % training set
data_ts.entries = proj_omega_sub(Us,Vs,data_ts.rows,data_ts.cols); % testing set

% If (not)interested in testing then (un)comment out the following line
%data_ts.entries =[]; data_ts.rows =[]; data_ts.cols=[];

%% Creating a grid of lambda values
lambda_threshold = 2*svds(sparse(data_ls.rows,data_ls.cols,data_ls.entries,d1,d2),1,'L'); 

lambda_max =  (1 - 1e-5)*lambda_threshold;% should be less than lambda_threshold for rank to be non-zero
lambda_min = 1e-3;
gamma = 0.85;



%% Setup initial model
model.d1 = d1;
model.d2 = d2;
model.p = 1; 
model.pmax = min(d1,d2);
model.lambda_max = lambda_max;
model.lambda_min = lambda_min;
model.gamma = gamma;

%% Functions
fun_set=@functions_nsym_polar_geometry; % File that specifies geometry of the search space UBV^T
fun_obj=@functions_matrix_completion_polar; % File that computes cost function, gradient of the matrix completion problem


%% Random Intialization
G = randn(model.d1, model.p); H = randn(model.d2, model.p);
[Qg Rg] = qr(G, 0);
[Qh Rh] = qr(H, 0);
[q1 b q2] = svd(Rg * Rh');

model.U = Qg * q1;
model.V = Qh * q2;
model.B = b;

%% Run algorithm
[model, preds_ts, infos] = lowrank_matrix_completion(optim_algo, data_ts, data_ls, model, params);




%% Plots
% Rank vs. Lambda plot

fs = 20;
figure('name', 'Rank fitting');
hl1 = plot(-log10(infos.Regpath_lambda), infos.Regpath_rank,'--*','Color','b','LineWidth',2);
ax1 = gca;

set(ax1,'FontSize',fs);

xlabel(ax1,'1/\lambda  in log scale','FontSize',fs);
ylabel(ax1,'Rank','FontSize',fs);

title('Rank fitting');
legend 'boxoff';


% Optimal cost vs. Lambda plot

fs = 20;

figure('name', 'Regularization path');

hl1 = semilogy(-log10(infos.Regpath_lambda), infos.Regpath_sol,'--*','Color','b','LineWidth',2);
ax1 = gca;

set(ax1,'FontSize',fs);

xlabel(ax1,'1/\lambda  in log scale','FontSize',fs);
ylabel(ax1,'Optimal cost','FontSize',fs);

title('Regularization path');
legend 'boxoff';


% Test error vs. Lambda plot

fs = 20;

figure('name', ' Test error Vs. Lambda');

hl1 = semilogy(-log10(infos.Regpath_lambda), infos.Regpath_preds_ts,'--*','Color','b','LineWidth',2);
ax1 = gca;

set(ax1,'FontSize',fs);

xlabel(ax1,'1/\lambda  in log scale','FontSize',fs);
ylabel(ax1,'Test error','FontSize',fs);

title('Matrix recovery with deacrasing \lambda');
legend 'boxoff';



% Relative duality gap plot

fs = 20;

figure('name', 'Optimality certificate');

hl1 = semilogy(-log10(infos.Regpath_lambda), abs(infos.Regpath_rel_duality_gap),'--*','Color','b','LineWidth',2);
ax1 = gca;

set(ax1,'FontSize',fs);

xlabel(ax1,'1/\lambda  in log scale','FontSize',fs);
ylabel(ax1,'Relative duality gap','FontSize',fs);

title('Optimality certificate');
legend 'boxoff';



% Relative comparison between Predictor-corrector and Warm restart

fs = 20;

figure('name', 'Predictor-corrector vs. Warm restart');
hl1 = plot(-log10(infos.Regpath_lambda(2:end)), (infos.Regpath_warm_preds(2:end) - infos.Regpath_sol(2:end)),'Color','r','LineWidth',2);
hold on;
hl1 = plot(-log10(infos.Regpath_lambda(2:end)), (infos.Regpath_predictor_preds(2:end) - infos.Regpath_sol(2:end)),'Color','b','LineWidth',2);

ax1 = gca;

set(ax1,'FontSize',fs);

xlabel(ax1,'1/\lambda  in log scale','FontSize',fs);
ylabel(ax1,'Inaccuracy in prediction','FontSize',fs);

title('Better prediction');
legend('Warm restart', 'Predictor-corrector');
legend 'boxoff';



