function [output cache] = functions_matrix_completion_polar(fname, model, searchDir, data, params, cache)
% Set of functions that are specific to matrix completion problem
%
% Total space = St(n,p)\times Cone(p)\times St(m,p)
% St(n,p): Stiefel manifold
% Cone(p): Positive definite matrices of size p\times p
%
% Usages:
%
% Cost function:                            f
% Gradient on the total space:              grad 
% Riemmanian Hessian on the total space:    hess 
% Dual variable:                            dual_var
% model is the current iterate
% eta is a search direction
% params is a cell array that contains {I,J,trueDists,EIJ} (see lowrank_dist_completion.m)
% cache is a cell array to reduce redundant computations
%
% Authors:
% Bamdev Mishra and Gilles Meyer
% {b.mishra,g.meyer}@ulg.ac.be

switch lower(fname),   
  
    case 'predict', % Compute predictions
      
      output = proj_omega_sub(model.U*model.B, model.V, data.rows, data.cols);
      
    case 'f', % Objective function
      
      preds = functions_matrix_completion_polar('predict', model, [], data);
      errors = (preds - data.entries);
      
      % Squared Error + trace norm 
      output = sum(errors.^2) + model.lambda*trace(model.B);
      
      
    case 'grad_f', % Compute gradient
      
      S = functions_matrix_completion_polar('dual_var', model, [], data);
        
      grad.U = S*model.V*model.B;
      grad.U = proj_tangent_space_stiefel(grad.U,model.U);
      
      grad.B = model.B*(symm(model.U'*S*model.V)) * model.B + model.lambda*model.B*model.B;
      
      
      grad.V = S'*model.U*model.B;
      grad.V = proj_tangent_space_stiefel(grad.V,model.V);
      
      % Cache the value of S and gradient 
      % for subsequent computation of the Hessian
      cache.S = S;
      cache.grad = grad;
              
      output = grad;
        
    case 'hessian', % Compute hessian

      if isfield(cache,'S'),
        S = cache.S;
      else
        S = functions_matrix_completion_polar('dual_var', model, [], data);
      end
      
      if isfield(cache,'grad'),
        grad = cache.grad;
      else
        grad = functions_matrix_completion_polar('grad_f', model, [], data);
      end

      s_star_u = proj_omega_sub(searchDir.U*model.B, model.V, data.rows, data.cols);
      s_star_b = proj_omega_sub(model.U*searchDir.B, model.V, data.rows, data.cols);
      s_star_v = proj_omega_sub(model.U*model.B, searchDir.V, data.rows, data.cols);
            
      S_star = sparse(data.rows, data.cols, 2 * (s_star_u + s_star_b + s_star_v), model.d1, model.d2);

      hessian.U = S*(model.V*searchDir.B + searchDir.V*model.B) + S_star*model.V*model.B;

      % Projection onto the tangent space to the Stiefel Manifold      
      hessian.U = proj_tangent_space_stiefel(hessian.U,model.U) - ...
                  proj_tangent_space_stiefel(searchDir.U*symm(model.U'*(S*model.V*model.B)),model.U);

      % Compute directional derivative      
      symmUSV = symm(model.U'*S*model.V);     
      
      hessian.B = searchDir.B*symmUSV*model.B  + ...
                     model.B*symmUSV*searchDir.B  + ...
                     model.B*symm(searchDir.U'*S*model.V + model.U'*S_star*model.V + model.U'*S*searchDir.V)*model.B + ...
                     model.lambda*model.B*searchDir.B + model.lambda*searchDir.B*model.B;
      
      % Correction for non-constant metric
      hessian.B = hessian.B - symm(searchDir.B*(model.B\grad.B));
      
      hessian.V = S'*(model.U*searchDir.B + searchDir.U*model.B) + S_star'*model.U*model.B;
      
      % Projection onto the tangent space to the Stiefel Manifold
      hessian.V = proj_tangent_space_stiefel(hessian.V,model.V) - ...
                  proj_tangent_space_stiefel(searchDir.V*symm(model.V'*(S'*model.U*model.B)),model.V);
      
      output = hessian;
        
    case 'dual_var', % Dual variable used to compute restart direction

        preds = functions_matrix_completion_polar('predict', model, [], data);
        errors = (preds - data.entries);
        output = sparse(data.rows, data.cols, 2 * errors, model.d1, model.d2);

    otherwise,
      error('Unknown function %s');
end

