function [model, preds_ts, infos] = tr_matrix_completion(data_ts, data_ls, model, params)
% [model, preds_ts, infos] = tr_matrix_completion(data_ts, data_ls, model, params)
%
% We obtain a local minimum (or a stationary point) of the non-convex problem
% minimize f(U, B, V) + \lambda Trace(B). 
% The search space is the quotietnt space. Details in the 
% "Low-rank optimization with trace norm penalty"
% Technical report (arXiv 1112.2318), 2011.
%
%
% The optimization algorithm is trust-region.
%
%
% Parameters
%
%   data_ls.rows               vector containing first indices of the points
%   data_ls.cols               vector containing second indices of the points
%   data_ls.entries            vector of known entries
%   data_ts                    structure array for testing (if not empty)
%   (model.U,model.B,model.V)  initial point
%   params                     structure array containing algorithm parameters
%                              (see default_arams for details)
%
%
% Output
%
%   model        structure array of the final point of rank model.p 
%   preds_ts     prediction for testing (computed only if data_ts is not empty)
%   infos        structure array with additional information
%
%
% Authors:
% Bamdev Mishra and Gilles Meyer
% {b.mishra,g.meyer}@ulg.ac.be  

    fun_set=@functions_nsym_polar_geometry; % Manifold related descriptions
    fun_obj=@functions_matrix_completion_polar; % Problem related ingredients
    
    t0 = tic;
    [model, infos] = trust_region(fun_set, fun_obj, data_ls, model, params);
    infos.time = toc(t0);   
    
    if ~isempty(data_ts),
        preds_ts = feval(fun_obj,'predict', model, [], data_ts);
    else
        preds_ts = [];
    end

end
